/*
 * $Id: servo1.c,v 1.2 2002/04/17 21:26:14 ldrolez Exp $
 *
 * Copyright (C) 2002 Ludovic Drolez
 *
 *   This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.       
 * 
 */

#include <io.h>
#include <interrupt.h>
#include <sig-avr.h>
#include <timer.h>

/* device ID (can be any char) */
#define ID '0'
/* timer 0 initial count */
/* 0 -> 25,6 us PWM resolution @ 10 Mhz */
/* 56 -> 20 us */
#define STARTCNT0 56


#undef ECHO

/* global variables used by PWM outputs */
volatile uint8_t pwm[8];
static uint8_t pwm_clk = 0;
static uint8_t pwm_cnt = 0;
volatile uint8_t pwm_config;

/* global variables for serial port */
static uint8_t recv_bytes = 0;
static char buffer[10];
volatile uint8_t ready = 0;

/* prototypes */
void dobuffer (void);

/*
 * Set default values
 */
void init(void)
{
  /* all outputs = pwm */
  pwm_config = 0xFF;
  /* all outputs low */
  memset (pwm, 0x00, 8);
}

/*
 * uart transmit complete isr
 */
INTERRUPT (SIG_UART_TRANS)
{
  ready = 1;
}

/*
 * Timer isr
 *
 * Takes 95 cycles (9,5us @10Mhz) with gcc 3.0.3 and -Os optimization
 */
SIGNAL (SIG_OVERFLOW0)
{
  uint8_t tpwm;

/* reload counter, +23 to compensate for all 
				   instructions executed between the timer
				   overflow and this reload */
  outp (STARTCNT0+23, TCNT0);	
  tpwm = 0;
  if (pwm_clk < pwm[0])
    tpwm |= 1;
  if (pwm_clk < pwm[1])
    tpwm |= 2;
  if (pwm_clk < pwm[2])
    tpwm |= 4;
  if (pwm_clk < pwm[3])
    tpwm |= 8;
  if (pwm_clk < pwm[4])
    tpwm |= 16;
  if (pwm_clk < pwm[5])
    tpwm |= 32;
  if (pwm_clk < pwm[6])
    tpwm |= 64;
  if (pwm_clk < pwm[7])
    tpwm |= 128;

  if (pwm_cnt & 1) {		/* in servo mode, only output 1 pulse / 2 */
    tpwm &= pwm_config;
    __asm__ __volatile__ ("nop"); /* improves timing */
  } else {
    __asm__ __volatile__ ("nop");
  }
  
  outp (tpwm, PORTB);		/* set PWM outputs */

  pwm_clk++;
  if (pwm_clk == 255)
    {
      pwm_clk = 0;		/* 255 steps not 256 */
      pwm_cnt++;
    }
}

/*
 * Convert an ASCII char to a 4 bits value
 */
uint8_t
quartet (char c)
{
  /* to upper case */
  if (c >= 'a')
    c = c - 32;
  if (c < 'A')
    return (c - '0');
  else
    return (c - 'A' + 10);
}

char
tobcd (uint8_t i)
{
  if (i > 9)
    return (i + 'A' - 10);
  else
    return (i + '0');
}

/*
 *
 */
void
pause (void)
{
  while (!ready);
  ready = 0;
}


/*
 * 
 */
void
receive_loop (void)
{
  uint8_t b;

  for (;;)
    {
      loop_until_bit_is_set (USR, 7);	/* wait for serial data */
      b = inp (UDR);		/* read it */
      buffer[recv_bytes++] = b;
#ifdef ECHO
      outp (b, UDR);
#endif
      if ((recv_bytes == 10) || (b < 32))
	{
	  dobuffer ();
	  recv_bytes = 0;
	}
    }
}

/*
 * 
 */
void
dobuffer (void)
{
  uint8_t tmpq1, pwmtoset;
  char *ptr;

  ptr = &buffer[0];
  if (*ptr++ != ID)
    return;			/* device ID not found, flush buffer */

  switch (*ptr++)
    {
    case 'P':			/* set PWM */
      pwmtoset = quartet (*ptr++);	/* get pwn # */
      ptr++;			/* get space */
      tmpq1 = quartet (*ptr++);	/* get value */
      pwm[pwmtoset] = tmpq1 * 16 + quartet (*ptr);	/* set pwm */
      break;
    case 'I':			/* input */
      tmpq1 = inp (PIND);
      outp (tobcd (tmpq1 >> 4), UDR);
      pause ();
      outp (tobcd (tmpq1 & 15), UDR);
      pause ();
      outp ('\n', UDR);
      break;
    case 'O':			/* output */
      ptr++;			/* get space */
      tmpq1 = quartet (*ptr++);	/* get value */
      outp ((tmpq1 * 16 + quartet (*ptr)) & 0xFC, PORTD);	/* set output */
      break;
    case 'C':			/* configure */
      switch (*ptr++)
	{
	case 'D':		/* set direction */
	  ptr++;		/* get space */
	  tmpq1 = quartet (*ptr++);	/* get value */
	  outp ((tmpq1 * 16 + quartet (*ptr)) & 0xFC, DDRD);	/* set direction bits */
	  break;
	case 'P':		/* set PWM mode */
	  ptr++;		/* get space */
	  tmpq1 = quartet (*ptr++);	/* get value */
	  pwm_config = (tmpq1 * 16 + quartet (*ptr));	/* set mode bits */
	  break;
	}
      break;
    }

}

/*
 *
 */
int
main (void)
{
  init();
  /* 10Mhz / 200 = 1 overflow every 20,0 us @ 10MHz */
  /* 255 steps = 5.1 ms */
  outp (0xff, DDRB);		/* portb = output */
  outp (0x00, DDRD);		/* portd = input */
  outp ((1 << TOIE0), TIMSK);
  outp (STARTCNT0, TCNT0);	/* timer start val */
  outp (CK, TCCR0);		/* prescale = 1 */

  outp ((1 << TXCIE) | (1 << RXEN) | (1 << TXEN), UCR);
  outp (64, UBRR);		/* baud rate =9600 */

  sei ();			/* enable interrupts */
  receive_loop ();
}
